package sun.text.normalizer;

import java.text.ParsePosition;
import java.util.Iterator;
import java.util.TreeSet;

public class UnicodeSet implements UnicodeMatcher
{
    private static final int LOW = 0;
    private static final int HIGH = 1114112;
    public static final int MIN_VALUE = 0;
    public static final int MAX_VALUE = 1114111;
    private int len;
    private int[] list;
    private int[] rangeList;
    private int[] buffer;
    TreeSet<String> strings;
    private String pat;
    private static final int START_EXTRA = 16;
    private static final int GROW_EXTRA = 16;
    private static UnicodeSet[] INCLUSIONS;
    static final VersionInfo NO_VERSION;
    public static final int IGNORE_SPACE = 1;
    
    public UnicodeSet() {
        this.strings = new TreeSet();
        this.pat = null;
        (this.list = new int[17])[this.len++] = 1114112;
    }
    
    public UnicodeSet(int n, int n2) {
        this();
        this.complement(n, n2);
    }
    
    public UnicodeSet(String s) {
        this();
        this.applyPattern(s, null, null, 1);
    }
    
    public UnicodeSet set(UnicodeSet set) {
        this.list = (int[])set.list.clone();
        this.len = set.len;
        this.pat = set.pat;
        this.strings = (TreeSet)set.strings.clone();
        return this;
    }
    
    public final UnicodeSet applyPattern(String s) {
        return this.applyPattern(s, null, null, 1);
    }
    
    private static void _appendToPat(StringBuffer sb, String s, boolean b) {
        for (int i = 0; i < s.length(); i += UTF16.getCharCount(i)) {
            _appendToPat(sb, UTF16.charAt(s, i), b);
        }
    }
    
    private static void _appendToPat(StringBuffer sb, int n, boolean b) {
        if (b && Utility.isUnprintable(n) && Utility.escapeUnprintable(sb, n)) {
            return;
        }
        switch (n) {
            case 36:
            case 38:
            case 45:
            case 58:
            case 91:
            case 92:
            case 93:
            case 94:
            case 123:
            case 125: {
                sb.append('\\');
                break;
            }
            default: {
                if (UCharacterProperty.isRuleWhiteSpace(n)) {
                    sb.append('\\');
                    break;
                }
                break;
            }
        }
        UTF16.append(sb, n);
        return;
    }
    
    private StringBuffer _toPattern(StringBuffer sb, boolean b) {
        if (this.pat != null) {
            int n = 0;
            int i = 0;
            while (i < this.pat.length()) {
                int char1 = UTF16.charAt(this.pat, i);
                i += UTF16.getCharCount(char1);
                if (b && Utility.isUnprintable(char1)) {
                    if (n % 2 == 1) {
                        sb.setLength(sb.length() - 1);
                    }
                    Utility.escapeUnprintable(sb, char1);
                    n = 0;
                }
                else {
                    UTF16.append(sb, char1);
                    if (char1 == 92) {
                        ++n;
                    }
                    else {
                        n = 0;
                    }
                }
            }
            return sb;
        }
        return this._generatePattern(sb, b, true);
    }
    
    public StringBuffer _generatePattern(StringBuffer sb, boolean b, boolean b2) {
        sb.append('[');
        int rangeCount = this.getRangeCount();
        if (rangeCount > 1 && this.getRangeStart(0) == 0 && this.getRangeEnd(rangeCount - 1) == 1114111) {
            sb.append('^');
            for (int j = 1; j < rangeCount; ++j) {
                int n = this.getRangeEnd(j - 1) + 1;
                int n2 = this.getRangeStart(j) - 1;
                _appendToPat(sb, n, b);
                if (n != n2) {
                    if (n + 1 != n2) {
                        sb.append('-');
                    }
                    _appendToPat(sb, n2, b);
                }
            }
        }
        else {
            for (int i = 0; i < rangeCount; ++i) {
                int rangeStart = this.getRangeStart(i);
                int rangeEnd = this.getRangeEnd(i);
                _appendToPat(sb, rangeStart, b);
                if (rangeStart != rangeEnd) {
                    if (rangeStart + 1 != rangeEnd) {
                        sb.append('-');
                    }
                    _appendToPat(sb, rangeEnd, b);
                }
            }
        }
        if (b2 && this.strings.size() > 0) {
            Iterator<String> iterator = this.strings.iterator();
            while (iterator.hasNext()) {
                sb.append('{');
                _appendToPat(sb, (String)iterator.next(), b);
                sb.append('}');
            }
        }
        return sb.append(']');
    }
    
    private UnicodeSet add_unchecked(int n, int n2) {
        if (n < 0 || n > 1114111) {
            throw new IllegalArgumentException("Invalid code point U+" + Utility.hex(n, 6));
        }
        if (n2 < 0 || n2 > 1114111) {
            throw new IllegalArgumentException("Invalid code point U+" + Utility.hex(n2, 6));
        }
        if (n < n2) {
            this.add(this.range(n, n2), 2, 0);
        }
        else if (n == n2) {
            this.add(n);
        }
        return this;
    }
    
    public final UnicodeSet add(int n) {
        return this.add_unchecked(n);
    }
    
    private final UnicodeSet add_unchecked(int n) {
        if (n < 0 || n > 1114111) {
            throw new IllegalArgumentException("Invalid code point U+" + Utility.hex(n, 6));
        }
        int codePoint = this.findCodePoint(n);
        if ((codePoint & 0x1) != 0x0) {
            return this;
        }
        if (n == this.list[codePoint] - 1) {
            if ((this.list[codePoint] = n) == 1114111) {
                this.ensureCapacity(this.len + 1);
                this.list[this.len++] = 1114112;
            }
            if (codePoint > 0 && n == this.list[codePoint - 1]) {
                System.arraycopy(this.list, codePoint + 1, this.list, codePoint - 1, this.len - codePoint - 1);
                this.len -= 2;
            }
        }
        else if (codePoint > 0 && n == this.list[codePoint - 1]) {
            int[] list2 = this.list;
            int n2 = codePoint - 1;
            ++list2[n2];
        }
        else {
            if (this.len + 2 > this.list.length) {
                int[] list = new int[this.len + 2 + 16];
                if (codePoint != 0) {
                    System.arraycopy(this.list, 0, list, 0, codePoint);
                }
                System.arraycopy(this.list, codePoint, list, codePoint + 2, this.len - codePoint);
                this.list = list;
            }
            else {
                System.arraycopy(this.list, codePoint, this.list, codePoint + 2, this.len - codePoint);
            }
            this.list[codePoint] = n;
            this.list[codePoint + 1] = n + 1;
            this.len += 2;
        }
        this.pat = null;
        return this;
    }
    
    public final UnicodeSet add(String e) {
        int singleCP = getSingleCP(e);
        if (singleCP < 0) {
            this.strings.add(e);
            this.pat = null;
            return this;
        }
        this.add_unchecked(singleCP, singleCP);
        return this;
    }
    
    private static int getSingleCP(String s) {
        if (s.length() < 1) {
            throw new IllegalArgumentException("Can't use zero-length strings in UnicodeSet");
        }
        if (s.length() > 2) {
            return -1;
        }
        if (s.length() == 1) {
            return s.charAt(0);
        }
        int char1 = UTF16.charAt(s, 0);
        if (char1 > 65535) {
            return char1;
        }
        return -1;
    }
    
    public UnicodeSet complement(int n, int n2) {
        if (n < 0 || n > 1114111) {
            throw new IllegalArgumentException("Invalid code point U+" + Utility.hex(n, 6));
        }
        if (n2 < 0 || n2 > 1114111) {
            throw new IllegalArgumentException("Invalid code point U+" + Utility.hex(n2, 6));
        }
        if (n <= n2) {
            this.xor(this.range(n, n2), 2, 0);
        }
        this.pat = null;
        return this;
    }
    
    public UnicodeSet complement() {
        if (this.list[0] == 0) {
            System.arraycopy(this.list, 1, this.list, 0, this.len - 1);
            --this.len;
        }
        else {
            this.ensureCapacity(this.len + 1);
            System.arraycopy(this.list, 0, this.list, 1, this.len);
            this.list[0] = 0;
            ++this.len;
        }
        this.pat = null;
        return this;
    }
    
    public boolean contains(int n) {
        if (n < 0 || n > 1114111) {
            throw new IllegalArgumentException("Invalid code point U+" + Utility.hex(n, 6));
        }
        return (this.findCodePoint(n) & 0x1) != 0x0;
    }
    
    private final int findCodePoint(int n) {
        if (n < this.list[0]) {
            return 0;
        }
        if (this.len >= 2 && n >= this.list[this.len - 2]) {
            return this.len - 1;
        }
        int n2 = 0;
        int n3 = this.len - 1;
        while (true) {
            int n4 = n2 + n3 >>> 1;
            if (n4 == n2) {
                break;
            }
            if (n < this.list[n4]) {
                n3 = n4;
            }
            else {
                n2 = n4;
            }
        }
        return n3;
    }
    
    public UnicodeSet addAll(UnicodeSet set) {
        this.add(set.list, set.len, 0);
        this.strings.addAll(set.strings);
        return this;
    }
    
    public UnicodeSet retainAll(UnicodeSet set) {
        this.retain(set.list, set.len, 0);
        this.strings.retainAll(set.strings);
        return this;
    }
    
    public UnicodeSet removeAll(UnicodeSet set) {
        this.retain(set.list, set.len, 2);
        this.strings.removeAll(set.strings);
        return this;
    }
    
    public UnicodeSet clear() {
        this.list[0] = 1114112;
        this.len = 1;
        this.pat = null;
        this.strings.clear();
        return this;
    }
    
    public int getRangeCount() {
        return this.len / 2;
    }
    
    public int getRangeStart(int n) {
        return this.list[n * 2];
    }
    
    public int getRangeEnd(int n) {
        return this.list[n * 2 + 1] - 1;
    }
    
    UnicodeSet applyPattern(String str, ParsePosition parsePosition, SymbolTable symbolTable, int n) {
        int n2 = parsePosition == null;
        if (!(n2 == 0)) {
            parsePosition = new ParsePosition(0);
        }
        StringBuffer o = new StringBuffer();
        RuleCharacterIterator o2 = new RuleCharacterIterator(str, symbolTable, parsePosition);
        this.applyPattern(o2, symbolTable, o, n);
        if (o2.inVariable()) {
            syntaxError(o2, "Extra chars in variable value");
        }
        this.pat = o.toString();
        if (!(n2 == 0)) {
            int i = parsePosition.getIndex();
            if ((n & 0x1) != 0x0) {
                i = Utility.skipWhitespace(str, i);
            }
            if (i != str.length()) {
                throw new IllegalArgumentException("Parse of \"" + str + "\" failed at " + i);
            }
        }
        return this;
    }
    
    void applyPattern(RuleCharacterIterator ruleCharacterIterator, SymbolTable symbolTable, StringBuffer sb, int n) {
        int n2 = 3;
        if ((n & 0x1) != 0x0) {
            n2 |= 0x4;
        }
        StringBuffer o = new StringBuffer();
        Object o2 = null;
        boolean n3 = false;
        Object o3 = null;
        Object pos = null;
        int n4 = 0;
        int n5 = 0;
        int n6 = 0;
        int c = 0;
        boolean n7 = false;
        this.clear();
        while (n6 != 2 && !ruleCharacterIterator.atEnd()) {
            int n8 = 0;
            int escaped = 0;
            o3 = null;
            int n9 = 0;
            if (resemblesPropertyPattern(ruleCharacterIterator, n2)) {
                n9 = 2;
            }
            else {
                pos = ruleCharacterIterator.getPos(pos);
                n8 = ruleCharacterIterator.next(n2);
                escaped = ruleCharacterIterator.isEscaped();
                if (n8 == 91 && escaped == 0) {
                    if (n6 == 1) {
                        ruleCharacterIterator.setPos(pos);
                        n9 = 1;
                    }
                    else {
                        n6 = 1;
                        o.append('[');
                        pos = ruleCharacterIterator.getPos(pos);
                        n8 = ruleCharacterIterator.next(n2);
                        int escaped2 = ruleCharacterIterator.isEscaped();
                        if (n8 == 94 && escaped2 == 0) {
                            n7 = true;
                            o.append('^');
                            pos = ruleCharacterIterator.getPos(pos);
                            n8 = ruleCharacterIterator.next(n2);
                            ruleCharacterIterator.isEscaped();
                        }
                        if (n8 == 45) {
                            escaped = 1;
                        }
                        else {
                            ruleCharacterIterator.setPos(pos);
                            continue;
                        }
                    }
                }
                else if (symbolTable != null) {
                    UnicodeMatcher lookupMatcher = symbolTable.lookupMatcher(n8);
                    if (lookupMatcher != null) {
                        try {
                            o3 = (UnicodeSet)lookupMatcher;
                            n9 = 3;
                        }
                        catch (ClassCastException ex) {
                            syntaxError(ruleCharacterIterator, "Syntax error");
                        }
                    }
                }
            }
            if (n9 != 0) {
                if (n4 == 1) {
                    if (c != '\0') {
                        syntaxError(ruleCharacterIterator, "Char expected after operator");
                    }
                    this.add_unchecked(n5, n5);
                    _appendToPat(o, n5, false);
                    c = '\0';
                }
                if (c == '-' || c == '&') {
                    o.append((char)c);
                }
                if (o3 == null) {
                    if (o3 == null) {
                        o3 = new UnicodeSet();
                    }
                }
                switch (n9) {
                    case 1: {
                        o3.applyPattern(ruleCharacterIterator, symbolTable, o, n);
                        break;
                    }
                    case 2: {
                        ruleCharacterIterator.skipIgnored(n2);
                        o3.applyPropertyPattern(ruleCharacterIterator, o, symbolTable);
                        break;
                    }
                    case 3: {
                        o3._toPattern(o, false);
                        break;
                    }
                }
                n3 = true;
                if (n6 == 0) {
                    this.set(o3);
                    n6 = 2;
                    break;
                }
                switch (c) {
                    case '-': {
                        this.removeAll(o3);
                        break;
                    }
                    case '&': {
                        this.retainAll(o3);
                        break;
                    }
                    case '\0': {
                        this.addAll(o3);
                        break;
                    }
                }
                c = '\0';
                n4 = 2;
            }
            else {
                if (n6 == 0) {
                    syntaxError(ruleCharacterIterator, "Missing '['");
                }
                if (escaped == 0) {
                    switch (n8) {
                        case 93: {
                            if (n4 == 1) {
                                this.add_unchecked(n5, n5);
                                _appendToPat(o, n5, false);
                            }
                            if (c == 45) {
                                this.add_unchecked(c, c);
                                o.append((char)c);
                            }
                            else if (c == 38) {
                                syntaxError(ruleCharacterIterator, "Trailing '&'");
                            }
                            o.append(']');
                            n6 = 2;
                            continue;
                        }
                        case 45: {
                            if (c == '\0') {
                                if (n4 != 0) {
                                    c = (char)n8;
                                    continue;
                                }
                                this.add_unchecked(n8, n8);
                                n8 = ruleCharacterIterator.next(n2);
                                int escaped3 = ruleCharacterIterator.isEscaped();
                                if (n8 == 93 && escaped3 == 0) {
                                    o.append("-]");
                                    n6 = 2;
                                    continue;
                                }
                            }
                            syntaxError(ruleCharacterIterator, "'-' not after char or set");
                            break;
                        }
                        case 38: {
                            if (n4 == 2 && c == '\0') {
                                c = (char)n8;
                                continue;
                            }
                            syntaxError(ruleCharacterIterator, "'&' not after set");
                            break;
                        }
                        case 94: {
                            syntaxError(ruleCharacterIterator, "'^' not after '['");
                            break;
                        }
                        case 123: {
                            if (c != '\0') {
                                syntaxError(ruleCharacterIterator, "Missing operand after operator");
                            }
                            if (n4 == 1) {
                                this.add_unchecked(n5, n5);
                                _appendToPat(o, n5, false);
                            }
                            n4 = 0;
                            if (o2 == null) {
                                o2 = new StringBuffer();
                            }
                            else {
                                o2.setLength(0);
                            }
                            boolean n10 = false;
                            while (!ruleCharacterIterator.atEnd()) {
                                int next = ruleCharacterIterator.next(n2);
                                int escaped4 = ruleCharacterIterator.isEscaped();
                                if (next == 125 && escaped4 == 0) {
                                    n10 = true;
                                    break;
                                }
                                UTF16.append(o2, next);
                            }
                            if (o2.length() < 1 || n10 == 0) {
                                syntaxError(ruleCharacterIterator, "Invalid multicharacter string");
                            }
                            this.add(o2.toString());
                            o.append('{');
                            _appendToPat(o, o2.toString(), false);
                            o.append('}');
                            continue;
                        }
                        case 36: {
                            pos = ruleCharacterIterator.getPos(pos);
                            n8 = ruleCharacterIterator.next(n2);
                            int escaped5 = ruleCharacterIterator.isEscaped();
                            int n11 = n8 == 93 && escaped5 == 0;
                            if (symbolTable == null && n11 == 0) {
                                n8 = 36;
                                ruleCharacterIterator.setPos(pos);
                                break;
                            }
                            if (n11 && c == '\0') {
                                if (n4 == 1) {
                                    this.add_unchecked(n5, n5);
                                    _appendToPat(o, n5, false);
                                }
                                this.add_unchecked(65535);
                                n3 = true;
                                o.append('$').append(']');
                                n6 = 2;
                                continue;
                            }
                            else {
                                syntaxError(ruleCharacterIterator, "Unquoted '$'");
                                break;
                            }
                        }
                    }
                }
                switch (n4) {
                    case 0: {
                        n4 = 1;
                        n5 = n8;
                        continue;
                    }
                    case 1: {
                        if (c == '-') {
                            if (n5 >= n8) {
                                syntaxError(ruleCharacterIterator, "Invalid range");
                            }
                            this.add_unchecked(n5, n8);
                            _appendToPat(o, n5, false);
                            o.append((char)c);
                            _appendToPat(o, n8, false);
                            c = (n4 = 0);
                            continue;
                        }
                        else {
                            this.add_unchecked(n5, n5);
                            _appendToPat(o, n5, false);
                            n5 = n8;
                            continue;
                        }
                    }
                    case 2: {
                        if (c != '\0') {
                            syntaxError(ruleCharacterIterator, "Set expected after operator");
                        }
                        n5 = n8;
                        n4 = 1;
                        continue;
                    }
                }
            }
        }
        if (n6 != 2) {
            syntaxError(ruleCharacterIterator, "Missing ']'");
        }
        ruleCharacterIterator.skipIgnored(n2);
        if (!(n7 == 0)) {
            this.complement();
        }
        if (!(n3 == 0)) {
            sb.append(o.toString());
        }
        else {
            this._generatePattern(sb, false, true);
        }
    }
    
    private static void syntaxError(RuleCharacterIterator ruleCharacterIterator, String str) {
        throw new IllegalArgumentException("Error: " + str + " at \"" + Utility.escape(ruleCharacterIterator.toString()) + 34);
    }
    
    private void ensureCapacity(int n) {
        if (n <= this.list.length) {
            return;
        }
        int[] list = new int[n + 16];
        System.arraycopy(this.list, 0, list, 0, this.len);
        this.list = list;
        return;
    }
    
    private void ensureBufferCapacity(int n) {
        if (this.buffer != null && n <= this.buffer.length) {
            return;
        }
        this.buffer = new int[n + 16];
        return;
    }
    
    private int[] range(int n, int n2) {
        if (this.rangeList == null) {
            this.rangeList = new int[] { n, n2 + 1, 1114112 };
        }
        else {
            this.rangeList[0] = n;
            this.rangeList[1] = n2 + 1;
        }
        return this.rangeList;
    }
    
    private UnicodeSet xor(int[] array, int n, int n2) {
        this.ensureBufferCapacity(this.len + n);
        int n3 = 0;
        int n4 = 0;
        int len = 0;
        int n5 = this.list[n3++];
        int n6;
        if (n2 == 1 || n2 == 2) {
            n6 = 0;
            if (array[n4] == 0) {
                ++n4;
                n6 = array[n4];
            }
        }
        else {
            n6 = array[n4++];
        }
        while (true) {
            if (n5 < n6) {
                this.buffer[len++] = n5;
                n5 = this.list[n3++];
            }
            else if (n6 < n5) {
                this.buffer[len++] = n6;
                n6 = array[n4++];
            }
            else if (n5 != 1114112) {
                n5 = this.list[n3++];
                n6 = array[n4++];
            }
            else {
                break;
            }
        }
        this.buffer[len++] = 1114112;
        this.len = len;
        int[] list = this.list;
        this.list = this.buffer;
        this.buffer = list;
        this.pat = null;
        return this;
    }
    
    private UnicodeSet add(int[] array, int n, int n2) {
        this.ensureBufferCapacity(this.len + n);
        int n3 = 0;
        int n4 = 0;
        int len = 0;
        int n5 = this.list[n3++];
        int max = array[n4++];
    Label_0620:
        while (true) {
            switch (n2) {
                case 0: {
                    if (n5 < max) {
                        if (len > 0 && n5 <= this.buffer[len - 1]) {
                            n5 = max(this.list[n3], this.buffer[--len]);
                        }
                        else {
                            this.buffer[len++] = n5;
                            n5 = this.list[n3];
                        }
                        ++n3;
                        n2 ^= 0x1;
                        continue;
                    }
                    else if (max < n5) {
                        if (len > 0 && max <= this.buffer[len - 1]) {
                            max = max(array[n4], this.buffer[--len]);
                        }
                        else {
                            this.buffer[len++] = max;
                            max = array[n4];
                        }
                        ++n4;
                        n2 ^= 0x2;
                        continue;
                    }
                    else {
                        if (n5 == 1114112) {
                            break Label_0620;
                        }
                        if (len > 0 && n5 <= this.buffer[len - 1]) {
                            n5 = max(this.list[n3], this.buffer[--len]);
                        }
                        else {
                            this.buffer[len++] = n5;
                            n5 = this.list[n3];
                        }
                        ++n3;
                        n2 ^= 0x1;
                        max = array[n4++];
                        n2 ^= 0x2;
                        continue;
                    }
                }
                case 3: {
                    if (max <= n5) {
                        if (n5 == 1114112) {
                            break Label_0620;
                        }
                        this.buffer[len++] = n5;
                    }
                    else {
                        if (max == 1114112) {
                            break Label_0620;
                        }
                        this.buffer[len++] = max;
                    }
                    n5 = this.list[n3++];
                    n2 ^= 0x1;
                    max = array[n4++];
                    n2 ^= 0x2;
                    continue;
                }
                case 1: {
                    if (n5 < max) {
                        this.buffer[len++] = n5;
                        n5 = this.list[n3++];
                        n2 ^= 0x1;
                        continue;
                    }
                    if (max < n5) {
                        max = array[n4++];
                        n2 ^= 0x2;
                        continue;
                    }
                    if (n5 == 1114112) {
                        break Label_0620;
                    }
                    n5 = this.list[n3++];
                    n2 ^= 0x1;
                    max = array[n4++];
                    n2 ^= 0x2;
                    continue;
                }
                case 2: {
                    if (max < n5) {
                        this.buffer[len++] = max;
                        max = array[n4++];
                        n2 ^= 0x2;
                        continue;
                    }
                    if (n5 < max) {
                        n5 = this.list[n3++];
                        n2 ^= 0x1;
                        continue;
                    }
                    if (n5 == 1114112) {
                        break Label_0620;
                    }
                    n5 = this.list[n3++];
                    n2 ^= 0x1;
                    max = array[n4++];
                    n2 ^= 0x2;
                    continue;
                }
            }
        }
        this.buffer[len++] = 1114112;
        this.len = len;
        int[] list = this.list;
        this.list = this.buffer;
        this.buffer = list;
        this.pat = null;
        return this;
    }
    
    private UnicodeSet retain(int[] array, int n, int n2) {
        this.ensureBufferCapacity(this.len + n);
        int n3 = 0;
        int n4 = 0;
        int len = 0;
        int n5 = this.list[n3++];
        int n6 = array[n4++];
    Label_0508:
        while (true) {
            switch (n2) {
                case 0: {
                    if (n5 < n6) {
                        n5 = this.list[n3++];
                        n2 ^= 0x1;
                        continue;
                    }
                    if (n6 < n5) {
                        n6 = array[n4++];
                        n2 ^= 0x2;
                        continue;
                    }
                    if (n5 == 1114112) {
                        break Label_0508;
                    }
                    this.buffer[len++] = n5;
                    n5 = this.list[n3++];
                    n2 ^= 0x1;
                    n6 = array[n4++];
                    n2 ^= 0x2;
                    continue;
                }
                case 3: {
                    if (n5 < n6) {
                        this.buffer[len++] = n5;
                        n5 = this.list[n3++];
                        n2 ^= 0x1;
                        continue;
                    }
                    if (n6 < n5) {
                        this.buffer[len++] = n6;
                        n6 = array[n4++];
                        n2 ^= 0x2;
                        continue;
                    }
                    if (n5 == 1114112) {
                        break Label_0508;
                    }
                    this.buffer[len++] = n5;
                    n5 = this.list[n3++];
                    n2 ^= 0x1;
                    n6 = array[n4++];
                    n2 ^= 0x2;
                    continue;
                }
                case 1: {
                    if (n5 < n6) {
                        n5 = this.list[n3++];
                        n2 ^= 0x1;
                        continue;
                    }
                    if (n6 < n5) {
                        this.buffer[len++] = n6;
                        n6 = array[n4++];
                        n2 ^= 0x2;
                        continue;
                    }
                    if (n5 == 1114112) {
                        break Label_0508;
                    }
                    n5 = this.list[n3++];
                    n2 ^= 0x1;
                    n6 = array[n4++];
                    n2 ^= 0x2;
                    continue;
                }
                case 2: {
                    if (n6 < n5) {
                        n6 = array[n4++];
                        n2 ^= 0x2;
                        continue;
                    }
                    if (n5 < n6) {
                        this.buffer[len++] = n5;
                        n5 = this.list[n3++];
                        n2 ^= 0x1;
                        continue;
                    }
                    if (n5 == 1114112) {
                        break Label_0508;
                    }
                    n5 = this.list[n3++];
                    n2 ^= 0x1;
                    n6 = array[n4++];
                    n2 ^= 0x2;
                    continue;
                }
            }
        }
        this.buffer[len++] = 1114112;
        this.len = len;
        int[] list = this.list;
        this.list = this.buffer;
        this.buffer = list;
        this.pat = null;
        return this;
    }
    
    private static final int max(int n, int n2) {
        return (n > n2) ? n : n2;
    }
    
    private static synchronized UnicodeSet getInclusions(int i) {
        if (UnicodeSet.INCLUSIONS == null) {
            UnicodeSet.INCLUSIONS = new UnicodeSet[9];
        }
        if (UnicodeSet.INCLUSIONS[i] == null) {
            UnicodeSet set = new UnicodeSet();
            switch (i) {
                case 2: {
                    UCharacterProperty.getInstance().upropsvec_addPropertyStarts(set);
                    UnicodeSet.INCLUSIONS[i] = set;
                    break;
                }
                default: {
                    throw new IllegalStateException("UnicodeSet.getInclusions(unknown src " + i + ")");
                }
            }
        }
        return UnicodeSet.INCLUSIONS[i];
    }
    
    private UnicodeSet applyFilter(UnicodeSet.Filter filter, int n) {
        this.clear();
        int n2 = -1;
        UnicodeSet inclusions = getInclusions(n);
        for (int rangeCount = inclusions.getRangeCount(), i = 0; i < rangeCount; ++i) {
            int j;
            int rangeEnd;
            for (j = inclusions.getRangeStart(i), rangeEnd = inclusions.getRangeEnd(i), ; j <= rangeEnd; ++j) {
                if (filter.contains(j)) {
                    if (n2 < 0) {
                        n2 = j;
                    }
                }
                else if (n2 >= 0) {
                    this.add_unchecked(n2, j - 1);
                    n2 = -1;
                }
            }
        }
        if (n2 >= 0) {
            this.add_unchecked(n2, 1114111);
            return this;
        }
        return this;
    }
    
    private static String mungeCharName(String s) {
        StringBuffer sb = new StringBuffer();
        int i = 0;
        while (i < s.length()) {
            int char1 = UTF16.charAt(s, i);
            i += UTF16.getCharCount(char1);
            if (UCharacterProperty.isRuleWhiteSpace(char1)) {
                if (sb.length() == 0) {
                    continue;
                }
                if (sb.charAt(sb.length() - 1) == ' ') {
                    continue;
                }
                char1 = 32;
            }
            UTF16.append(sb, char1);
        }
        if (sb.length() != 0 && sb.charAt(sb.length() - 1) == ' ') {
            sb.setLength(sb.length() - 1);
        }
        return sb.toString();
    }
    
    public UnicodeSet applyPropertyAlias(String str, String s, SymbolTable symbolTable) {
        if (s.length() > 0 && str.equals("Age")) {
            this.applyFilter(new UnicodeSet.VersionFilter(VersionInfo.getInstance(mungeCharName(s))), 2);
            return this;
        }
        throw new IllegalArgumentException("Unsupported property: " + str);
    }
    
    private static boolean resemblesPropertyPattern(RuleCharacterIterator ruleCharacterIterator, int n) {
        boolean n2 = false;
        n &= 0xFFFFFFFD;
        Object pos = ruleCharacterIterator.getPos(null);
        int next = ruleCharacterIterator.next(n);
        if (next == 91 || next == 92) {
            int next2 = ruleCharacterIterator.next(n & 0xFFFFFFFB);
            n2 = ((next == 91) ? (next2 == 58) : (next2 == 78 || next2 == 112 || next2 == 80));
        }
        ruleCharacterIterator.setPos(pos);
        return n2;
    }
    
    private UnicodeSet applyPropertyPattern(String s, ParsePosition parsePosition, SymbolTable symbolTable) {
        int index = parsePosition.getIndex();
        if (index + 5 > s.length()) {
            return null;
        }
        boolean n = false;
        boolean n2 = false;
        int n3 = 0;
        int beginIndex;
        if (s.regionMatches(index, "[:", 0, 2)) {
            n = true;
            beginIndex = Utility.skipWhitespace(s, index + 2);
            if (beginIndex < s.length() && s.charAt(beginIndex) == '^') {
                ++beginIndex;
                n3 = 1;
            }
        }
        else if (s.regionMatches(true, index, "\\p", 0, 2) || s.regionMatches(index, "\\N", 0, 2)) {
            int char1 = s.charAt(index + 1);
            n3 = (char1 == 'P');
            n2 = (char1 == 'N');
            beginIndex = Utility.skipWhitespace(s, index + 2);
            if (beginIndex == s.length() || s.charAt(beginIndex++) != '{') {
                return null;
            }
        }
        else {
            return null;
        }
        int index2 = s.indexOf(n ? ":]" : "}", beginIndex);
        if (index2 < 0) {
            return null;
        }
        int index3 = s.indexOf(61, beginIndex);
        String s2;
        String substring;
        if (index3 >= 0 && index3 < index2 && n2 == 0) {
            s2 = s.substring(beginIndex, index3);
            substring = s.substring(index3 + 1, index2);
        }
        else {
            s2 = s.substring(beginIndex, index2);
            substring = "";
            if (n2 != 0) {
                substring = s2;
                s2 = "na";
            }
        }
        this.applyPropertyAlias(s2, substring, symbolTable);
        if (n3 != 0) {
            this.complement();
        }
        parsePosition.setIndex(index2 + (n ? 2 : 1));
        return this;
    }
    
    private void applyPropertyPattern(RuleCharacterIterator ruleCharacterIterator, StringBuffer sb, SymbolTable symbolTable) {
        String lookahead = ruleCharacterIterator.lookahead();
        ParsePosition parsePosition = new ParsePosition(0);
        this.applyPropertyPattern(lookahead, parsePosition, symbolTable);
        if (parsePosition.getIndex() == 0) {
            syntaxError(ruleCharacterIterator, "Invalid property pattern");
        }
        ruleCharacterIterator.jumpahead(parsePosition.getIndex());
        sb.append(lookahead.substring(0, parsePosition.getIndex()));
    }
    
    static {
        UnicodeSet.INCLUSIONS = null;
        NO_VERSION = VersionInfo.getInstance(0, 0, 0, 0);
    }
    
    private static class VersionFilter implements UnicodeSet.Filter
    {
        VersionInfo version;
        
        VersionFilter(VersionInfo version) {
            this.version = version;
        }
        
        @Override
        public boolean contains(int n) {
            VersionInfo age = UCharacter.getAge(n);
            return age != UnicodeSet.NO_VERSION && age.compareTo(this.version) <= 0;
        }
    }
    
    private interface Filter
    {
        public abstract boolean contains(int p0);
    }
}
